/* Ekiga -- A VoIP and Video-Conferencing application
 * Copyright (C) 2000-2009 Damien Sandras <dsandras@seconix.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 *
 * Ekiga is licensed under the GPL license and as a special exception,
 * you have permission to link or otherwise combine this program with the
 * programs OPAL, OpenH323 and PWLIB, and distribute the combination,
 * without applying the requirements of the GNU GPL to the OPAL, OpenH323
 * and PWLIB programs, as long as you do follow the requirements of the
 * GNU GPL for all the rest of the software thus combined.
 */


/*
 *                         codecsbox.c  -  description
 *                         -------------------------------
 *   begin                : Sat Sep 2 2006
 *   copyright            : (C) 2000-2006 by Damien Sandras
 *   description          : Contains a codecs box widget
 *
 */

#include "codecsbox.h"

#include <sstream>

#include <glib/gi18n.h>

#include "call-core.h"
#include "codec-description.h"
#include "ekiga-settings.h"

/* Columns for the codecs page */
enum {

  COLUMN_CODEC_ACTIVE,
  COLUMN_CODEC_NAME,
  COLUMN_CODEC_DISPLAY_NAME,
  COLUMN_CODEC_INFO,
  COLUMN_CODEC_SELECTABLE,
  COLUMN_CODEC_NUMBER
};

struct _CodecsBoxPrivate
{
  Ekiga::Call::StreamType type;
  GtkWidget *codecs_list;
  boost::shared_ptr<Ekiga::Settings> audio_settings;
  boost::shared_ptr<Ekiga::Settings> video_settings;
  boost::shared_ptr<Ekiga::CallCore> call_core;
  GmApplication *app;
};

enum { TYPE = 1 };

G_DEFINE_TYPE (CodecsBox, codecs_box, GTK_TYPE_BOX);

/* Static functions */
static void codecs_box_set_codecs (CodecsBox *self,
                                   const Ekiga::CodecList & list);


/* GTK+ Callbacks */
static void codec_toggled_cb (GtkCellRendererToggle *cell,
                              gchar *path_str,
                              gpointer data);

static void codec_moved_cb (GtkWidget *widget,
                            gpointer data);

static std::list<std::string> codecs_box_to_list (CodecsBox *self);

static void codecs_box_build (CodecsBox *self);

static void codecs_box_class_init (CodecsBoxClass* klass);

static void codecs_box_init (CodecsBox *);

static void codecs_box_dispose (GObject *obj);

static void codecs_box_get_property (GObject *obj,
                                     guint prop_id,
                                     GValue *value,
                                     GParamSpec *spec);

static void codecs_box_set_property (GObject *obj,
                                     guint prop_id,
                                     const GValue *value,
                                     GParamSpec *spec);

static void
codecs_box_set_codecs (CodecsBox *self,
                       const Ekiga::CodecList & list)
{
  GtkTreeSelection *selection = NULL;
  GtkTreeModel *model = NULL;
  GtkTreeIter iter;

  gchar *selected_codec = NULL;
  unsigned select_rate = 0;
  bool selected = false;

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (self->priv->codecs_list));
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (self->priv->codecs_list));

  if (gtk_tree_selection_get_selected (selection, &model, &iter))
    gtk_tree_model_get (GTK_TREE_MODEL (model), &iter,
                        COLUMN_CODEC_NAME, &selected_codec,
                        COLUMN_CODEC_INFO, &select_rate, -1);
  gtk_list_store_clear (GTK_LIST_STORE (model));

  for (Ekiga::CodecList::const_iterator itr = list.begin ();
       itr != list.end ();
       itr++) {

    if ((self->priv->type == Ekiga::Call::Audio && (*itr).audio)
        || (self->priv->type == Ekiga::Call::Video && !(*itr).audio)) {

      gtk_list_store_append (GTK_LIST_STORE (model), &iter);
      gtk_list_store_set (GTK_LIST_STORE (model), &iter,
                          COLUMN_CODEC_ACTIVE, (*itr).active,
                          COLUMN_CODEC_NAME, (*itr).name.c_str (),
                          COLUMN_CODEC_DISPLAY_NAME, (*itr).display_name.c_str (),
                          COLUMN_CODEC_INFO, (*itr).display_info.c_str (),
                          COLUMN_CODEC_SELECTABLE, "true",
                          -1);

      if (selected_codec && (*itr).name == selected_codec) {

        selected = true;
        gtk_tree_selection_select_iter (selection, &iter);
      }
    }
  }

  if (!selected && gtk_tree_model_get_iter_first (GTK_TREE_MODEL (model), &iter))
    gtk_tree_selection_select_iter (selection, &iter);

  g_free (selected_codec);
}


static void
codec_toggled_cb (G_GNUC_UNUSED GtkCellRendererToggle *cell,
                  gchar *path_str,
                  gpointer data)
{
  CodecsBox *self = NULL;

  GtkTreeModel *model = NULL;
  GtkTreePath *path = NULL;
  GtkTreeIter iter;

  std::list<std::string> list;

  gboolean fixed = FALSE;

  self = CODECS_BOX (data);

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (self->priv->codecs_list));
  path = gtk_tree_path_new_from_string (path_str);

  /* Update the tree model */
  gtk_tree_model_get_iter (model, &iter, path);
  gtk_tree_model_get (model, &iter, COLUMN_CODEC_ACTIVE, &fixed, -1);
  gtk_list_store_set (GTK_LIST_STORE (model), &iter,
                      COLUMN_CODEC_ACTIVE, fixed^1, -1);
  gtk_tree_path_free (path);

  /* Update the gmconf key */
  list = codecs_box_to_list (self);
  if (self->priv->type == Ekiga::Call::Audio)
    self->priv->audio_settings->set_string_list ("media-list", list);
  else if (self->priv->type == Ekiga::Call::Video)
    self->priv->video_settings->set_string_list ("media-list", list);
}


static void
codec_moved_cb (GtkWidget *widget,
                gpointer data)
{
  CodecsBox *self = NULL;

  GtkTreeIter iter;
  GtkTreeIter *iter2 = NULL;
  GtkTreeModel *model = NULL;
  GtkTreeSelection *selection = NULL;
  GtkTreePath *tree_path = NULL;

  std::list<std::string> list;

  gchar *path_str = NULL;

  self = CODECS_BOX (data);

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (self->priv->codecs_list));
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (self->priv->codecs_list));
  if (!gtk_tree_selection_get_selected (GTK_TREE_SELECTION (selection),
                                        NULL, &iter))
    return;

  /* Update the tree view */
  iter2 = gtk_tree_iter_copy (&iter);
  path_str = gtk_tree_model_get_string_from_iter (GTK_TREE_MODEL (model), &iter);
  tree_path = gtk_tree_path_new_from_string (path_str);
  if (!g_strcmp0 ((gchar *) g_object_get_data (G_OBJECT (widget), "operation"), "up"))
    gtk_tree_path_prev (tree_path);
  else
    gtk_tree_path_next (tree_path);

  gtk_tree_model_get_iter (GTK_TREE_MODEL (model), &iter, tree_path);
  if (gtk_list_store_iter_is_valid (GTK_LIST_STORE (model), &iter)
      && gtk_list_store_iter_is_valid (GTK_LIST_STORE (model), iter2))
    gtk_list_store_swap (GTK_LIST_STORE (model), &iter, iter2);

  /* Scroll to the new position */
  gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (self->priv->codecs_list),
                                tree_path, NULL, FALSE, 0, 0);

  gtk_tree_path_free (tree_path);
  gtk_tree_iter_free (iter2);
  g_free (path_str);

  /* Update the key */
  list = codecs_box_to_list (self);
  if (self->priv->type == Ekiga::Call::Audio)
    self->priv->audio_settings->set_string_list ("media-list", list);
  else if (self->priv->type == Ekiga::Call::Video)
    self->priv->video_settings->set_string_list ("media-list", list);
}


static std::list<std::string>
codecs_box_to_list (CodecsBox *self)
{
  GtkTreeModel *model = NULL;
  GtkTreeIter iter;

  gchar *name = NULL;
  gboolean active = false;

  std::list<std::string> list;

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (self->priv->codecs_list));
  if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (model), &iter)) {

    do {
      std::stringstream val;

      gtk_tree_model_get (model, &iter,
                          COLUMN_CODEC_ACTIVE, &active,
                          COLUMN_CODEC_NAME, &name,
                          -1);

      val << name << ":" << (active ? "1" : "0");
      list.push_back (val.str ());
      g_free (name);

    } while (gtk_tree_model_iter_next (GTK_TREE_MODEL (model), &iter));
  }

  return list;
}


static void
codecs_box_build (CodecsBox *self)
{
  GtkWidget *image = NULL;
  GtkWidget *scroll_window = NULL;
  GtkWidget *button = NULL;

  GtkWidget *buttons_vbox = NULL;

  GtkListStore *list_store = NULL;
  GtkCellRenderer *renderer = NULL;
  GtkTreeViewColumn *column = NULL;

  gtk_box_set_spacing (GTK_BOX (self), 6);
  gtk_box_set_homogeneous (GTK_BOX (self), FALSE);

  list_store = gtk_list_store_new (COLUMN_CODEC_NUMBER,
                                   G_TYPE_BOOLEAN,
                                   G_TYPE_STRING,
                                   G_TYPE_STRING,
                                   G_TYPE_STRING,
                                   G_TYPE_BOOLEAN,
                                   G_TYPE_BOOLEAN);

  gtk_tree_view_set_reorderable (GTK_TREE_VIEW (self->priv->codecs_list), TRUE);
  gtk_tree_view_set_search_column (GTK_TREE_VIEW (self->priv->codecs_list),0);
  gtk_tree_view_set_model (GTK_TREE_VIEW (self->priv->codecs_list),
                           GTK_TREE_MODEL (list_store));
  gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (self->priv->codecs_list), FALSE);

  /* Set all Colums */
  renderer = gtk_cell_renderer_toggle_new ();
  column = gtk_tree_view_column_new_with_attributes (NULL,
                                                     renderer,
                                                     "active",
                                                     COLUMN_CODEC_ACTIVE,
                                                     NULL);
  gtk_tree_view_column_add_attribute (column, renderer,
                                      "activatable", COLUMN_CODEC_SELECTABLE);
  g_object_set (renderer, "xalign", 0.0, "xpad", 6, NULL);
  gtk_tree_view_append_column (GTK_TREE_VIEW (self->priv->codecs_list), column);
  g_signal_connect (renderer, "toggled", G_CALLBACK (codec_toggled_cb), (gpointer) self);

  renderer = gtk_cell_renderer_text_new ();
  column = gtk_tree_view_column_new ();
  gtk_tree_view_append_column (GTK_TREE_VIEW (self->priv->codecs_list), column);
  gtk_tree_view_column_pack_start (column, renderer, TRUE);
  gtk_tree_view_column_add_attribute (column, renderer, "text", COLUMN_CODEC_DISPLAY_NAME);
  g_object_set (G_OBJECT (renderer), "weight", PANGO_WEIGHT_BOLD,
                "xalign", 0.0, NULL);

  renderer = gtk_cell_renderer_text_new ();
  gtk_tree_view_column_pack_start (column, renderer, TRUE);
  gtk_tree_view_column_add_attribute (column, renderer, "text", COLUMN_CODEC_INFO);
  g_object_set (G_OBJECT (renderer), "style", PANGO_STYLE_ITALIC,
                "ellipsize", PANGO_ELLIPSIZE_END, "width-chars", 30,
                "scale", 0.85, "xalign", 0, "xpad", 6, NULL);

  scroll_window = gtk_scrolled_window_new (FALSE, FALSE);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll_window),
                                  GTK_POLICY_NEVER,
                                  GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scroll_window),
                                       GTK_SHADOW_IN);
  gtk_widget_set_size_request (scroll_window, -1, 130);
  gtk_container_add (GTK_CONTAINER (scroll_window),
                     GTK_WIDGET (self->priv->codecs_list));
  gtk_box_pack_start (GTK_BOX (self), scroll_window, TRUE, TRUE, 0);


  /* The buttons */
  buttons_vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
  gtk_widget_set_vexpand (GTK_WIDGET (buttons_vbox), TRUE);

  image = gtk_image_new_from_icon_name ("go-up-symbolic", GTK_ICON_SIZE_MENU);
  button = gtk_button_new ();
  gtk_container_add (GTK_CONTAINER (button), image);
  gtk_widget_set_tooltip_text (button, _("Move selected codec priority upwards"));
  gtk_box_pack_start (GTK_BOX (buttons_vbox), button, FALSE, FALSE, 0);
  g_object_set_data (G_OBJECT (button), "operation", (gpointer) "up");
  gtk_widget_set_valign (GTK_WIDGET (button), GTK_ALIGN_START);
  g_signal_connect (button, "clicked",
                    G_CALLBACK (codec_moved_cb),
                    (gpointer) self);

  image = gtk_image_new_from_icon_name ("go-down-symbolic", GTK_ICON_SIZE_MENU);
  button = gtk_button_new ();
  gtk_container_add (GTK_CONTAINER (button), image);
  gtk_widget_set_tooltip_text (button, _("Move selected codec priority downwards"));
  gtk_box_pack_start (GTK_BOX (buttons_vbox), button, FALSE, FALSE, 0);
  g_object_set_data (G_OBJECT (button), "operation", (gpointer) "down");
  gtk_widget_set_valign (GTK_WIDGET (button), GTK_ALIGN_END);
  g_signal_connect (button, "clicked",
                    G_CALLBACK (codec_moved_cb),
                    (gpointer) self);

  gtk_box_pack_start (GTK_BOX (self), buttons_vbox, FALSE, FALSE, 0);

  gtk_widget_set_vexpand (GTK_WIDGET (self), TRUE);
  gtk_widget_set_hexpand (GTK_WIDGET (self), TRUE);
  gtk_widget_show_all (GTK_WIDGET (self));
}



static void
codecs_box_class_init (CodecsBoxClass* klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GParamSpec *spec = NULL;

  gobject_class->dispose = codecs_box_dispose;
  gobject_class->get_property = codecs_box_get_property;
  gobject_class->set_property = codecs_box_set_property;

  spec = g_param_spec_int ("type", "Type", "Type",
                           0, 1, 0, (GParamFlags) G_PARAM_READWRITE);
  g_object_class_install_property (gobject_class, TYPE, spec);
}


static void
codecs_box_init (G_GNUC_UNUSED CodecsBox *self)
{
  /* Nothing to do here, unfortunately */
}


static void
codecs_box_dispose (GObject *obj)
{
  CodecsBox *self = NULL;

  self = CODECS_BOX (obj);

  self->priv->codecs_list = NULL;

  G_OBJECT_CLASS (codecs_box_parent_class)->dispose (obj);
}

static void
codecs_box_get_property (GObject *obj,
                         guint prop_id,
                         GValue *value,
                         GParamSpec *spec)
{
  CodecsBox *self = CODECS_BOX (obj);

  switch (prop_id) {

  case TYPE:
    g_value_set_int (value, self->priv->type);
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, spec);
    break;
  }
}


static void
codecs_box_set_property (GObject *obj,
                         guint prop_id,
                         const GValue *value,
                         GParamSpec *spec)
{
  CodecsBox *self = CODECS_BOX (obj);
  std::list<std::string> list;

  switch (prop_id) {

  case TYPE:
    self->priv->type = (Ekiga::Call::StreamType) g_value_get_int (value);
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, spec);
    break;
  }

  if (self->priv->type == Ekiga::Call::Audio)
    list = self->priv->audio_settings->get_string_list ("media-list");
  else if (self->priv->type == Ekiga::Call::Video)
    list = self->priv->video_settings->get_string_list ("media-list");
}


/* public api */

GtkWidget *
codecs_box_new_with_type (GmApplication *app,
                          Ekiga::Call::StreamType type)
{
  CodecsBox *self = CODECS_BOX (g_object_new (CODECS_BOX_TYPE, NULL));

  Ekiga::ServiceCore& core = gm_application_get_core (app);

  self->priv = new _CodecsBoxPrivate ();
  self->priv->app = app;
  self->priv->call_core = core.get<Ekiga::CallCore> ("call-core");
  self->priv->type = type;
  self->priv->audio_settings = boost::shared_ptr<Ekiga::Settings> (new Ekiga::Settings (AUDIO_CODECS_SCHEMA));
  self->priv->video_settings = boost::shared_ptr<Ekiga::Settings> (new Ekiga::Settings (VIDEO_CODECS_SCHEMA));
  self->priv->codecs_list = gtk_tree_view_new ();

  codecs_box_build (self);

  codecs_box_set_codecs (self, self->priv->call_core->get_codecs ());

  return GTK_WIDGET (self);
}
