/*
 * Ekiga -- A VoIP and Video-Conferencing application
 * Copyright (C) 2000-2009 Damien Sandras <dsandras@seconix.com>

 * This program is free software; you can  redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version. This program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Ekiga is licensed under the GPL license and as a special exception, you
 * have permission to link or otherwise combine this program with the
 * programs OPAL, OpenH323 and PWLIB, and distribute the combination, without
 * applying the requirements of the GNU GPL to the OPAL, OpenH323 and PWLIB
 * programs, as long as you do follow the requirements of the GNU GPL for all
 * the rest of the software thus combined.
 */


/*
 *                         videoinput-core.cpp  -  description
 *                         ------------------------------------------
 *   begin                : written in 2008 by Matthias Schneider
 *   copyright            : (c) 2008 by Matthias Schneider
 *   description          : declaration of the interface of a videoinput core.
 *                          A videoinput core manages VideoInputManagers.
 *
 */

#if DEBUG
#include <typeinfo>
#include <iostream>
#endif

#include <glib/gi18n.h>

#include "config.h"

#include "videoinput-core.h"
#include "videooutput-manager.h"
#include "videoinput-manager.h"

using namespace Ekiga;


VideoInputCore::VideoPreviewManager::VideoPreviewManager (VideoInputCore& _videoinput_core, boost::shared_ptr<VideoOutputCore> _videooutput_core)
: PThread (1000, NoAutoDeleteThread, HighestPriority, "VideoPreviewManager"),
    videoinput_core (_videoinput_core),
  videooutput_core (_videooutput_core)
{
  width = 176;
  height = 144;;
  pause_thread = true;
  end_thread = false;
  frame = NULL;
  // Since windows does not like to restart a thread that
  // was never started, we do so here
  this->Resume ();
}

VideoInputCore::VideoPreviewManager::~VideoPreviewManager ()
{
  quit ();

#if DEBUG
  std::cout << "Destroyed object of type " << typeid(*this).name () << std::endl;
#endif
}

void VideoInputCore::VideoPreviewManager::quit ()
{
  {
    PWaitAndSignal q(exit_mutex);
    end_thread = true;
  }

  {
    PWaitAndSignal m(thread_mutex);
    stop ();
  }
}

void VideoInputCore::VideoPreviewManager::start (unsigned _width, unsigned _height)
{
  PTRACE(4, "PreviewManager\tStarting Preview");

  {
    PWaitAndSignal c(capture_mutex);
    width = _width;
    height = _height;
    pause_thread = false;
  }
  {
    PWaitAndSignal c(frame_mutex);
    frame = (char*) malloc (unsigned (width * height * 3 / 2));
  }

  videooutput_core->start();
}

void VideoInputCore::VideoPreviewManager::stop ()
{
  PTRACE(4, "PreviewManager\tStopping Preview");

  videooutput_core->stop();

  {
    PWaitAndSignal c(capture_mutex);
    if (pause_thread)
      return;
    pause_thread = true;
  }

  {
    PWaitAndSignal c(frame_mutex);
    if (frame)
      free (frame);
  }
}

void VideoInputCore::VideoPreviewManager::Main ()
{
  PWaitAndSignal m(thread_mutex);
  bool exit = end_thread;
  bool capture;

  while (!exit) {

    {
      PWaitAndSignal c(capture_mutex);
      capture = !pause_thread;
    }
    if (capture) {
      {
        PWaitAndSignal c(frame_mutex);
        if (frame) {
          videoinput_core.get_frame_data(frame);
          videooutput_core->set_frame_data(frame, width, height, VideoOutputManager::LOCAL, 1);
        }
      }
    }
    {
       PWaitAndSignal q(exit_mutex);
       exit = end_thread;
    }
    Current()->Sleep (50);
  }
}

void VideoInputCore::VideoPreviewManager::Terminate ()
{
  quit ();
}


VideoInputCore::VideoInputCore (Ekiga::ServiceCore & _core,
                                boost::shared_ptr<VideoOutputCore> _videooutput_core)
  : core(_core)
{
  PWaitAndSignal m_var(core_mutex);
  PWaitAndSignal m_set(settings_mutex);

  preview_manager = new VideoPreviewManager (*this, _videooutput_core);

  preview_config.active = false;
  preview_config.width = 176;
  preview_config.height = 144;
  preview_config.fps = 30;

  stream_config.active = false;
  stream_config.width = 176;
  stream_config.height = 144;
  stream_config.fps = 30;

  current_settings.brightness = 0;
  current_settings.whiteness = 0;
  current_settings.colour = 0;
  current_settings.contrast = 0;

  desired_settings.brightness = 0;
  desired_settings.whiteness = 0;
  desired_settings.colour = 0;
  desired_settings.contrast = 0;

  current_manager = NULL;
  notification_core = core.get<Ekiga::NotificationCore> ("notification-core");

  device_settings = new Settings (VIDEO_DEVICES_SCHEMA);
  device_settings->changed.connect (boost::bind (&VideoInputCore::setup, this, _1));

  video_codecs_settings = new Settings (VIDEO_CODECS_SCHEMA);
  video_codecs_settings->changed.connect (boost::bind (&VideoInputCore::setup, this, _1));
}


VideoInputCore::~VideoInputCore ()
{
  delete preview_manager;

  PWaitAndSignal m(core_mutex);

  for (std::set<VideoInputManager *>::iterator iter = managers.begin ();
       iter != managers.end ();
       iter++)
    delete (*iter);

  managers.clear();

  delete device_settings;
  delete video_codecs_settings;
}


void VideoInputCore::setup (std::string setting)
{
  GSettings* settings = device_settings->get_g_settings ();
  GSettings* codecs_settings = video_codecs_settings->get_g_settings ();
  VideoInputDevice device;

  /* Get device settings */
  if (setting == "any" || setting == "input-device" || setting == "format" || setting == "channel") {
    gchar *device_string = g_settings_get_string (settings, "input-device");
    unsigned video_format = g_settings_get_enum (settings, "format");
    unsigned channel = g_settings_get_int (settings, "channel");
    device.SetFromString (device_string);

    set_device (device, channel, (VideoInputFormat) video_format);
    g_free (device_string);
  }

  /* Size and framerate */
  if (setting == "any" || setting == "size" || setting == "max-frame-rate") {
    unsigned size = g_settings_get_enum (settings, "size");
    unsigned max_frame_rate = g_settings_get_int (codecs_settings, "max-frame-rate");
    if (size >= NB_VIDEO_SIZES) {
      PTRACE(1, "VidInputCore\t" << "size out of range, ajusting to 0");
      size = 0;
    }
    set_preview_config (VideoSizes[size].width,
                        VideoSizes[size].height,
                        max_frame_rate);
  }
}


void VideoInputCore::add_manager (VideoInputManager &manager)
{
  managers.insert (&manager);
  manager_added (manager);

  manager.device_opened.connect (boost::bind (&VideoInputCore::on_device_opened, this, _1, _2, &manager));
  manager.device_closed.connect (boost::bind (&VideoInputCore::on_device_closed, this, _1, &manager));
  manager.device_error.connect (boost::bind (&VideoInputCore::on_device_error, this, _1, _2, &manager));
}


void VideoInputCore::visit_managers (boost::function1<bool, VideoInputManager &> visitor) const
{
  PWaitAndSignal m(core_mutex);
  bool go_on = true;

  for (std::set<VideoInputManager *>::const_iterator iter = managers.begin ();
       iter != managers.end () && go_on;
       iter++)
      go_on = visitor (*(*iter));
}

void VideoInputCore::get_devices (std::vector <std::string> & devices)
{
  std::vector <VideoInputDevice> d;
  get_devices (d);

  devices.clear ();

  for (std::vector<VideoInputDevice>::iterator iter = d.begin ();
       iter != d.end ();
       ++iter)
    devices.push_back (iter->GetString ());
}

void VideoInputCore::get_devices (std::vector <VideoInputDevice> & devices)
{
  PWaitAndSignal m(core_mutex);

  devices.clear();

  for (std::set<VideoInputManager *>::iterator iter = managers.begin ();
       iter != managers.end ();
       iter++)
    (*iter)->get_devices (devices);

#if PTRACING
  for (std::vector<VideoInputDevice>::iterator iter = devices.begin ();
       iter != devices.end ();
       iter++) {
    PTRACE(4, "VidInputCore\tDetected Device: " << *iter);
  }
#endif
}

void VideoInputCore::set_device(const VideoInputDevice & _device, int channel, VideoInputFormat format)
{
  PWaitAndSignal m(core_mutex);
  GSettings* settings = device_settings->get_g_settings ();
  VideoInputDevice device, device1, device2;

  /* Check if device exists */
  std::vector <VideoInputDevice> devices;
  bool found = false;
  bool found_preferred1 = false;
  bool found_preferred2 = false;

  get_devices (devices);
  for (std::vector<VideoInputDevice>::iterator it = devices.begin ();
       it < devices.end ();
       it++) {
    if (*it == _device) {
      found = true;
      break;
    } else if (it->GetString ().find (VIDEO_INPUT_PREFERRED_DEVICE_TYPE1) != string::npos
               && it->GetString ().find (VIDEO_INPUT_PREFERRED_DEVICE_SOURCE1) != string::npos) {
      found_preferred1 = true;
      device1 = *it;  // do not break, we still hope to find _device
    } else if (it->GetString ().find (VIDEO_INPUT_PREFERRED_DEVICE_TYPE2) != string::npos
               && it->GetString ().find (VIDEO_INPUT_PREFERRED_DEVICE_SOURCE2) != string::npos) {
      found_preferred2 = true;
      device2 = *it;  // do not break, we still hope to find _device or preferred1
    }
  }

  if (found)
    device = _device;
  else if (found_preferred1)
    device = device1;
  else if (found_preferred2)
    device = device2;
  else if (!devices.empty ())
    device = *devices.begin();
  else {
    device.type = VIDEO_INPUT_FALLBACK_DEVICE_TYPE;
    device.source = VIDEO_INPUT_FALLBACK_DEVICE_SOURCE;
    device.name = VIDEO_INPUT_FALLBACK_DEVICE_NAME;
  }

  PTRACE(4, "VidInputCoreConf\tSet video input device to " << device);

  if (format >= VI_FORMAT_MAX) {
    PTRACE(1, "VidInputCoreConf\tformat out of range, adjusting to 3");
    format = (VideoInputFormat) 3;
  }

  if (!found)
    g_settings_set_string (settings, "input-device", device.GetString ().c_str ());
  else
    internal_set_device (device, channel, format);
}

void VideoInputCore::add_device (const std::string & source, const std::string & device_name, unsigned capabilities, HalManager* /*manager*/)
{
  PTRACE(4, "VidInputCore\tAdding device " << device_name);
  PWaitAndSignal m(core_mutex);

  VideoInputDevice device;
  for (std::set<VideoInputManager *>::iterator iter = managers.begin ();
       iter != managers.end ();
       iter++) {
    if ((*iter)->has_device (source, device_name, capabilities, device)) {

      device_added (device);

      boost::shared_ptr<Ekiga::Notification> notif (new Ekiga::Notification (Ekiga::Notification::Info, _("New Video Input Device"), device.GetString (), _("Use It"), boost::bind (&VideoInputCore::on_set_device, (VideoInputCore*) this, device)));
      notification_core->push_notification (notif);
    }
  }
}

void VideoInputCore::remove_device (const std::string & source, const std::string & device_name, unsigned capabilities, HalManager* /*manager*/)
{
  PTRACE(4, "VidInputCore\tRemoving device " << device_name);
  PWaitAndSignal m(core_mutex);

  VideoInputDevice device;
  for (std::set<VideoInputManager *>::iterator iter = managers.begin ();
       iter != managers.end ();
       iter++) {
     if ((*iter)->has_device (source, device_name, capabilities, device)) {
       if ( (current_device == device) && (preview_config.active || stream_config.active) ) {

            VideoInputDevice new_device;
            new_device.type   = VIDEO_INPUT_FALLBACK_DEVICE_TYPE;
            new_device.source = VIDEO_INPUT_FALLBACK_DEVICE_SOURCE;
            new_device.name   = VIDEO_INPUT_FALLBACK_DEVICE_NAME;
            internal_set_device(new_device, current_channel, current_format);
       }

       device_removed(device, current_device == device);
     }
  }
}

void VideoInputCore::set_preview_config (unsigned width, unsigned height, unsigned fps)
{
  PWaitAndSignal m(core_mutex);

  if ( fps < 1 || fps > 30 ) {
    PTRACE(1, "VidInputCore\tmax-frame-rate out of range, ajusting to 30");
    fps = 30;
  }

  VideoDeviceConfig new_preview_config(width, height, fps);

  PTRACE(4, "VidInputCore\tSetting new preview config: " << new_preview_config);
  // There is only one state where we have to reopen the preview device:
  // we have preview enabled, no stream is active and some value has changed
  if ( ( preview_config.active && !stream_config.active) &&
       ( preview_config        !=  new_preview_config) )
  {
    preview_manager->stop();
    internal_close();

    internal_open(new_preview_config.width, new_preview_config.height, new_preview_config.fps);
    preview_manager->start(new_preview_config.width, new_preview_config.height);
  }

  preview_config = new_preview_config;
}


void VideoInputCore::start_preview ()
{
  PWaitAndSignal m(core_mutex);

  PTRACE(4, "VidInputCore\tStarting preview " << preview_config);
  if (!preview_config.active && !stream_config.active) {
    internal_open(preview_config.width, preview_config.height, preview_config.fps);
    preview_manager->start(preview_config.width, preview_config.height);
  }

  preview_config.active = true;
}

void VideoInputCore::stop_preview ()
{
  PWaitAndSignal m(core_mutex);

  PTRACE(4, "VidInputCore\tStopping Preview");
  if (preview_config.active && !stream_config.active) {
    preview_manager->stop();
    internal_close();
  }

  preview_config.active = false;
}

void VideoInputCore::set_stream_config (unsigned width, unsigned height, unsigned fps)
{
  PWaitAndSignal m(core_mutex);

  VideoDeviceConfig new_stream_config(width, height, fps);
  PTRACE(4, "VidInputCore\tSetting new stream config: " << new_stream_config);

  // We do not support switching of framerate or resolution within a stream
  // since many endpoints will probably have problems with that. Also, it would add
  // a lot of complexity due to the capabilities exchange. Thus these values will
  // not be used until the next start_stream.

  if (!stream_config.active)
    stream_config = new_stream_config;
}

void VideoInputCore::start_stream ()
{
  PWaitAndSignal m(core_mutex);

  PTRACE(4, "VidInputCore\tStarting stream " << stream_config);
  if (preview_config.active && !stream_config.active) {
    preview_manager->stop();
    if ( preview_config != stream_config )
    {
      internal_close();
      internal_open(stream_config.width, stream_config.height, stream_config.fps);
    }
  }

  if (!preview_config.active && !stream_config.active) {
    internal_open(stream_config.width, stream_config.height, stream_config.fps);
  }

  stream_config.active = true;
}

void VideoInputCore::stop_stream ()
{
  PWaitAndSignal m(core_mutex);

  PTRACE(4, "VidInputCore\tStopping Stream");
  if (preview_config.active && stream_config.active) {
    if ( preview_config != stream_config )
    {
      internal_close();
      internal_open(preview_config.width, preview_config.height, preview_config.fps);
    }
    preview_manager->start(preview_config.width, preview_config.height);
  }

  if (!preview_config.active && stream_config.active) {
    internal_close();
  }

  stream_config.active = false;
}

void VideoInputCore::get_frame_data (char *data)
{
  if (current_manager) {
    if (!current_manager->get_frame_data(data)) {

      PWaitAndSignal m(core_mutex);
      internal_close();

      if (preview_config.active && !stream_config.active)
        internal_open(preview_config.width, preview_config.height, preview_config.fps);

      if (stream_config.active)
        internal_open(stream_config.width, stream_config.height, stream_config.fps);

      if (current_manager)
        current_manager->get_frame_data(data); // the default device must always return true
    }
    internal_apply_settings();
  }
}

void VideoInputCore::set_colour (unsigned colour)
{
  PWaitAndSignal m(settings_mutex);
  desired_settings.colour = colour;
}

void VideoInputCore::set_brightness (unsigned brightness)
{
  PWaitAndSignal m(settings_mutex);
  desired_settings.brightness = brightness;
}

void VideoInputCore::set_whiteness  (unsigned whiteness)
{
  PWaitAndSignal m(settings_mutex);
  desired_settings.whiteness = whiteness;
}

void VideoInputCore::set_contrast   (unsigned contrast)
{
  PWaitAndSignal m(settings_mutex);
  desired_settings.contrast = contrast;
}

void VideoInputCore::on_set_device (const VideoInputDevice & device)
{
  g_settings_set_string (device_settings->get_g_settings (), "input-device", device.GetString ().c_str ());
}

void VideoInputCore::on_device_opened (VideoInputDevice device,
                                     VideoInputSettings settings,
                                     VideoInputManager *manager)
{
  device_opened (*manager, device, settings);
}

void VideoInputCore::on_device_closed (VideoInputDevice device, VideoInputManager *manager)
{
  device_closed (*manager, device);
}

void VideoInputCore::on_device_error (VideoInputDevice device, VideoInputErrorCodes error_code, VideoInputManager *manager)
{
  device_error (*manager, device, error_code);
}

void VideoInputCore::internal_set_device(const VideoInputDevice & device, int channel, VideoInputFormat format)
{
  PTRACE(4, "VidInputCore\tSetting device: " << device);

  if (preview_config.active && !stream_config.active)
    preview_manager->stop();

  if (preview_config.active || stream_config.active)
    internal_close();

  internal_set_manager (device, channel, format);

  if (preview_config.active && !stream_config.active) {
    internal_open(preview_config.width, preview_config.height, preview_config.fps);
    preview_manager->start(preview_config.width,preview_config.height);
  }

  if (stream_config.active)
    internal_open(stream_config.width, stream_config.height, stream_config.fps);
}

void VideoInputCore::internal_set_manager (const VideoInputDevice & device, int channel, VideoInputFormat format)
{
  current_manager = NULL;
  for (std::set<VideoInputManager *>::iterator iter = managers.begin ();
       iter != managers.end ();
       iter++) {
     if ((*iter)->set_device (device, channel, format)) {
       current_manager = (*iter);
     }
  }

  // If the desired manager could not be found,
  // we se the default device. The default device
  // MUST ALWAYS be loaded and openable
  if (current_manager) {
    current_device  = device;
  }
  else {
    PTRACE(1, "VidInputCore\tTried to set unexisting device " << device);
    internal_set_fallback();
  }

  current_channel = channel;
  current_format  = format;
}


void VideoInputCore::internal_set_fallback ()
{
  current_device.type   = VIDEO_INPUT_FALLBACK_DEVICE_TYPE;
  current_device.source = VIDEO_INPUT_FALLBACK_DEVICE_SOURCE;
  current_device.name   = VIDEO_INPUT_FALLBACK_DEVICE_NAME;
  PTRACE(3, "VidInputCore\tFalling back to " << current_device);

  internal_set_manager(current_device, current_channel, current_format);
}

void VideoInputCore::internal_open (unsigned width, unsigned height, unsigned fps)
{
  PTRACE(4, "VidInputCore\tOpening device with " << width << "x" << height << "/" << fps );

  if (current_manager && !current_manager->open(width, height, fps)) {

    internal_set_fallback();
    if (current_manager)
      current_manager->open(width, height, fps);
  }
}

void VideoInputCore::internal_close()
{
  PTRACE(4, "VidInputCore\tClosing current device");
  if (current_manager)
    current_manager->close();
}

void VideoInputCore::internal_apply_settings()
{
  PWaitAndSignal m_set(settings_mutex);

  if (desired_settings.colour != current_settings.colour) {
    current_manager->set_colour (desired_settings.colour);
    current_settings.colour = desired_settings.colour;
  }

  if (desired_settings.brightness != current_settings.brightness) {
    current_manager->set_brightness (desired_settings.brightness);
    current_settings.brightness = desired_settings.brightness;
  }

  if (desired_settings.whiteness != current_settings.whiteness) {
    current_manager->set_whiteness (desired_settings.whiteness);
    current_settings.whiteness = desired_settings.whiteness;
  }

  if (desired_settings.contrast != current_settings.contrast) {
    current_manager->set_contrast (desired_settings.contrast);
    current_settings.contrast = desired_settings.contrast;
  }
}
